@@ -39,10 +39,14 @@
 #include <linux/if_ether.h>
 #include <linux/filter.h>
 #include <net/if_arp.h>
+#include <errno.h>
+#include <sys/time.h>
 
 #include "io.h"
 #include "util.h"
 #include "private.h"
+#include "time.h"
+#include "time-private.h"
 #include "dhcp-private.h"
 
 struct dhcp_default_transport {
@@ -110,8 +114,21 @@ static bool _dhcp_default_transport_read_handler(struct l_io *io,
 	ssize_t len;
 	struct dhcp_packet *p;
 	uint16_t c;
+	struct sockaddr_ll saddr;
+	uint64_t timestamp = 0;
+	struct cmsghdr *cmsg;
+	struct iovec iov = { .iov_base = buf, .iov_len = sizeof(buf) };
+	struct msghdr msg = {};
+	unsigned char control[32 + CMSG_SPACE(sizeof(struct timeval))];
 
-	len = read(fd, buf, sizeof(buf));
+	msg.msg_name = &saddr;
+	msg.msg_namelen = sizeof(saddr);
+	msg.msg_iov = &iov;
+	msg.msg_iovlen = 1;
+	msg.msg_control = control;
+	msg.msg_controllen = sizeof(control);
+
+	len = recvmsg(fd, &msg, 0);
 	if (len < 0)
 		return false;
 
@@ -147,8 +164,31 @@ static bool _dhcp_default_transport_read_handler(struct l_io *io,
 
 	len -= sizeof(struct udphdr) - sizeof(struct iphdr);
 
-	if (transport->super.rx_cb)
-		transport->super.rx_cb(&p->dhcp, len, transport->super.rx_data);
+	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg != NULL;
+					cmsg = CMSG_NXTHDR(&msg, cmsg)) {
+		if (cmsg->cmsg_level == SOL_SOCKET &&
+				cmsg->cmsg_type == SCM_TIMESTAMP &&
+				cmsg->cmsg_len ==
+				CMSG_LEN(sizeof(struct timeval))) {
+			const struct timeval *tv = (void *) CMSG_DATA(cmsg);
+
+			timestamp = _time_realtime_to_boottime(tv);
+		}
+	}
+
+	if (!timestamp)
+		timestamp = l_time_now();
+
+	if (transport->super.rx_cb) {
+		const uint8_t *src_mac = NULL;
+
+		if (msg.msg_namelen >= sizeof(saddr) &&
+				saddr.sll_halen == ETH_ALEN)
+			src_mac = saddr.sll_addr;
+
+		transport->super.rx_cb(&p->dhcp, len, transport->super.rx_data,
+					src_mac, timestamp);
+	}
 
 	return true;
 }
@@ -168,8 +208,8 @@ static void dhcp_set_ip_udp_headers(struct iphdr *ip, struct udphdr *udp,
 	ip->tot_len = L_CPU_TO_BE16(len + sizeof(*ip) + sizeof(*udp));
 
 	ip->protocol = IPPROTO_UDP;
-	ip->saddr = L_CPU_TO_BE32(saddr);
-	ip->daddr = L_CPU_TO_BE32(daddr);
+	ip->saddr = saddr;
+	ip->daddr = daddr;
 
 	udp->source = L_CPU_TO_BE16(sport);
 	udp->dest = L_CPU_TO_BE16(dport);
@@ -194,9 +234,10 @@ static void dhcp_set_ip_udp_headers(struct iphdr *ip, struct udphdr *udp,
 	ip->check = _dhcp_checksumv(iov, 1);
 }
 
-static int _dhcp_default_transport_broadcast(struct dhcp_transport *s,
+static int _dhcp_default_transport_l2_send(struct dhcp_transport *s,
 						uint32_t saddr, uint16_t sport,
 						uint32_t daddr, uint16_t dport,
+						const uint8_t *dest_mac,
 						const void *data, size_t len)
 {
 	struct dhcp_default_transport *transport =
@@ -221,7 +262,10 @@ static int _dhcp_default_transport_broadcast(struct dhcp_transport *s,
 	addr.sll_protocol = htons(ETH_P_IP);
 	addr.sll_ifindex = s->ifindex;
 	addr.sll_halen = ETH_ALEN;
-	memset(addr.sll_addr, 0xff, ETH_ALEN);
+	if (!dest_mac)
+		memset(addr.sll_addr, 0xff, ETH_ALEN);
+	else
+		memcpy(addr.sll_addr, dest_mac, ETH_ALEN);
 
 	memset(&msg, 0, sizeof(msg));
 	msg.msg_name = &addr;
@@ -319,6 +363,9 @@ static int _dhcp_default_transport_bind(struct dhcp_transport *s,
 	if (!transport->io)
 		return -EIO;
 
+	if (transport->udp_fd >= 0)
+		return 0;
+
 	fd = kernel_udp_socket_open(transport->ifname, saddr, transport->port);
 	if (fd < 0)
 		return fd;
@@ -342,18 +389,8 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 		BPF_STMT(BPF_RET + BPF_K, 0),
 		/* A <- IP version + Header length */
 		BPF_STMT(BPF_LD + BPF_B + BPF_ABS, 0),
-		/* A <- A & 0xf0 (Mask off version */
-		BPF_STMT(BPF_ALU + BPF_AND + BPF_K, 0xf0),
-		/* A == IPVERSION (shifted left 4) ? */
-		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, IPVERSION << 4, 1, 0),
-		/* ignore */
-		BPF_STMT(BPF_RET + BPF_K, 0),
-		/* A <- IP version + Header length */
-		BPF_STMT(BPF_LD + BPF_B + BPF_ABS, 0),
-		/* A <- A & 0x0f (Mask off IP Header Length */
-		BPF_STMT(BPF_ALU + BPF_AND + BPF_K, 0x0f),
-		/* A == 5 ? */
-		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, 5, 1, 0),
+		/* IP version == IPVERSION && Header length == 5 ? */
+		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, (IPVERSION << 4) | 5, 1, 0),
 		/* ignore */
 		BPF_STMT(BPF_RET + BPF_K, 0),
 		/* A <- IP protocol */
@@ -382,9 +419,11 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 		/* A <- DHCP op */
 		BPF_STMT(BPF_LD + BPF_B + BPF_ABS,
 				offsetof(struct dhcp_packet, dhcp.op)),
-		/* op == BOOTREPLY ? */
+		/* op == (BOOTREPLY | BOOTREQUEST) ? */
 		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K,
-				DHCP_OP_CODE_BOOTREPLY, 1, 0),
+				(port == DHCP_PORT_CLIENT) ?
+				DHCP_OP_CODE_BOOTREPLY :
+				DHCP_OP_CODE_BOOTREQUEST, 1, 0),
 		/* ignore */
 		BPF_STMT(BPF_RET + BPF_K, 0),
 		/* A <- DHCP header type */
@@ -394,6 +433,10 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ARPHRD_ETHER, 1, 0),
 		/* ignore */
 		BPF_STMT(BPF_RET + BPF_K, 0),
+		/* A <- UDP destination port */
+		BPF_STMT(BPF_LD + BPF_H + BPF_ABS,
+				offsetof(struct dhcp_packet, udp.dest)),
+		BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, DHCP_PORT_SERVER, 3, 0),
 		/* A <- client identifier */
 		BPF_STMT(BPF_LD + BPF_W + BPF_ABS,
 				offsetof(struct dhcp_packet, dhcp.xid)),
@@ -422,6 +465,7 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 		.len = L_ARRAY_SIZE(filter),
 		.filter = filter
 	};
+	int one = 1;
 
 	s = socket(AF_PACKET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (s < 0)
@@ -431,6 +475,9 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 						&fprog, sizeof(fprog)) < 0)
 		goto error;
 
+	if (setsockopt(s, SOL_SOCKET, SO_TIMESTAMP, &one, sizeof(one)) < 0)
+		goto error;
+
 	memset(&addr, 0, sizeof(addr));
 	addr.sll_family = AF_PACKET;
 	addr.sll_protocol = htons(ETH_P_IP);
@@ -462,6 +509,11 @@ static int _dhcp_default_transport_open(struct dhcp_transport *s, uint32_t xid)
 		return fd;
 
 	transport->io = l_io_new(fd);
+	if (!transport->io) {
+		close(fd);
+		return -EMFILE;
+	}
+
 	l_io_set_close_on_destroy(transport->io, true);
 	l_io_set_read_handler(transport->io,
 					_dhcp_default_transport_read_handler,
@@ -495,11 +547,12 @@ struct dhcp_transport *_dhcp_default_transport_new(uint32_t ifindex,
 	transport->super.bind = _dhcp_default_transport_bind;
 	transport->super.close = _dhcp_default_transport_close;
 	transport->super.send = _dhcp_default_transport_send;
-	transport->super.broadcast = _dhcp_default_transport_broadcast;
+	transport->super.l2_send = _dhcp_default_transport_l2_send;
 
 	transport->super.ifindex = ifindex;
 	l_strlcpy(transport->ifname, ifname, IFNAMSIZ);
 	transport->port = port;
+	transport->udp_fd = -1;
 
 	return &transport->super;
 }
